# --------------------------------------------------------
# Hessian Method Functions
# --------------------------------------------------------
hessianmethod <- function(funfcn,pars,y,delta,print=TRUE){
  # Applies the Hessian method
  # funfcn - function which return negative log-likelihood
  # pars - parameter values at which the Hessian is evaluated
  # y - data to be passed into the log-likelihood
  # delta - error used in calculating Hessian and cut-off
  # suggested value delta = 0.00001
  
  cutoff <- delta*length(pars) # cut-off uses delta*p, where p is no. pars
  h <- do.call("hessian",list(funfcn,pars,y,delta)) # Finds Hessian matrix
  E <- eigen(h) # Calculates eigenvalues
  standeigenvalues <- abs(E$values)/max(abs(E$values)) # Finds standardised eigenvalues
  # Find number of estimable parameters
  # Number of parameters with eigenvalues below the cut-off
  noestpars <- 0
  for (i in 1:length(pars)) {
    if (standeigenvalues[i] > cutoff) {
      noestpars <- noestpars + 1
    } 
  }
  if (print) {
     # Prints whether model is parameter redundant or not
     if (min(standeigenvalues) < cutoff) {
        cat("model is non-identifiable or parameter redundant")
     } 
     else {
        cat("model is identifiable or not parameter redundant")
     }
     cat("\n")  
     cat('smallest standardised eigenvalue',min(standeigenvalues)) # prints smallest eigenvalue
     cat("\n")  
     cat('number of estimable parameters',noestpars) # prints number of parameters
  }
  result <- list(standeigenvalues=standeigenvalues,noestpars=noestpars,E=E)
  return(result)
}

hessian <- function(funfcn,x,y,delta){
  # function to calcuate the Hessian of funfcn at x 
  t <- length(x)
  h <- matrix(0, t, t)
  Dx <- delta*diag(t)
  for (i in 1:t) {
    for (j in 1:i) {
      h[i,j] <- (do.call("funfcn",list(x+Dx[i,]+Dx[j,],y))- do.call("funfcn",list(x+Dx[i,]-Dx[j,],y))- do.call("funfcn",list(x-Dx[i,]+Dx[j,],y)) + do.call("funfcn",list(x-Dx[i,]-Dx[j,],y)))/(4*delta^2)
      h[j,i] <- h[i,j]
    }
  }
  return(h)
}

# -----------------------------------------------------------------------
# Logit and Inverse Logit (expit) functions
# -----------------------------------------------------------------------
expit <- function(xval) { 
  1/(1+exp(-xval))
}
logit <- function(xval) {
  log(xval/(1-xval))
}


# ------------------------------------------------------------------------------
# Log-Likelihood function for Lapwing example with parameters phi1, phia and rho
# ------------------------------------------------------------------------------
statespacelik <- function(pars,y){
  # This function returns the negative log-likelihood state-space model
  phi1 <- expit(pars[1])
  phia <- expit(pars[2])
  rho <- exp(pars[3])
  sigmasq <- 6
  TT <- matrix(c(0,rho*phi1,phia,phia), nrow=2, ncol=2, byrow=TRUE) 
  Z <- matrix(c(0,1), nrow=1, ncol=2, byrow=TRUE) 
  V <- eigen(TT)
  vv <- t(t(V$vec[ ,1]))
  a1 <- matrix(c(800,1000), nrow=2, ncol=1, byrow=TRUE)
  P1 <- matrix(c(0,0,0,0), nrow=2, ncol=2, byrow=TRUE)
  Qq <- matrix(c(0,rho*phi1,phia*(1-phia),phia*(1-phia)), nrow=2, ncol=2, byrow=TRUE)
  H <- sigmasq
  
  l <-0
  for (i in 1:length(y)) { 
    Q <- diag((Qq %*% a1)[ ,1])
    a <- TT %*% a1
    P <- TT %*% P1 %*% t(TT) + Q
    v <- y[i]-Z %*% a
    FF <- Z %*% P %*% t(Z)+H
    M <- P %*% t(Z)
    a1 <- a + M %*% solve(FF)  %*% v
    P1 <- P - M %*% solve(FF) %*% t(M)

    l <- l - 1/2 * ( log(det(FF) ) + t(v) %*% solve(FF) %*% v)
  }

  return(-l)
}

# -----------------------------------------------------------------------------------------
# Log-Likelihood function for reparameterised Lapwing example with parameters beta and phia
# -----------------------------------------------------------------------------------------
statespacelik2 <- function(pars,y){
  # This function returns the negative log-likelihood reparameterised state-space model
  beta <- exp(pars[1])
  phia <- expit(pars[2])
  sigmasq <- 6
  TT <- matrix(c(0,beta,phia,phia), nrow=2, ncol=2, byrow=TRUE) 
  Z <- matrix(c(0,1), nrow=1, ncol=2, byrow=TRUE) 
  V <- eigen(TT)
  vv <- t(t(V$vec[ ,1]))
  a1 <- matrix(c(800,1000), nrow=2, ncol=1, byrow=TRUE)
  P1 <- matrix(c(0,0,0,0), nrow=2, ncol=2, byrow=TRUE)
  Qq <- matrix(c(0,beta,phia*(1-phia),phia*(1-phia)), nrow=2, ncol=2, byrow=TRUE)
  H <- sigmasq
  
  l <-0
  for (i in 1:length(y)) { 
    Q <- diag((Qq %*% a1)[ ,1])
    a <- TT %*% a1
    P <- TT %*% P1 %*% t(TT) + Q
    v <- y[i]-Z %*% a
    FF <- Z %*% P %*% t(Z)+H
    M <- P %*% t(Z)
    a1 <- a + M %*% solve(FF)  %*% v
    P1 <- P - M %*% solve(FF) %*% t(M)
    
    l <- l - 1/2 * ( log(det(FF) ) + t(v) %*% solve(FF) %*% v)
  }
  
  return(-l)
}

#------------------------------------------------------------------
# Lapwing data
#------------------------------------------------------------------
y <-c(1092,1100,1234,1461,1570,1820,1391,1508,1541,1631)

#------------------------------------------------------------------
# The Hessian method for the introductory example 
#------------------------------------------------------------------

#Starting values for maximisation algorithm
phi1s <- 0.5
phias <- 0.6
rhos <- 1.5
inpars <- c(logit(phi1s),logit(phias),log(rhos))

# First find the maximum likeliood estimates. 
# Note phi1 and phia are maximised on the logit scale and rho
# is maximised on the log scale. 
maxlik <- optim(inpars,statespacelik,y=y,method="L-BFGS-B",hessian=TRUE)
print(maxlik)
pars <- c(expit(maxlik$par[1:2]),exp(maxlik$par[3]))
pars
# Although not needed the standard errors can also be found
H <- maxlik$hessian
var <- solve(H)
se <- sqrt(diag(var))*(c(pars[1:2]*(1-pars[1:2]), pars[3]))
se

# Then apply the Hessian method using the function hessianmethod
# This needs as inputs: a log-likelihood function, the MLEs, the data,
# the precision value for Hessian calculation
results <- hessianmethod(statespacelik,maxlik$par,y,0.00001,print=TRUE)
results

# Try changing the precision value 0.00001 and print to FALSE

#------------------------------------------------------------------
# The Hessian method for the reparameterised introductory example 
#------------------------------------------------------------------

#Starting values for maximisation algorithum
betas <- 0.8
phias <- 0.6
inpars <- c(log(betas),logit(phias))

# First find the maximum likeliood estimates. 
# Note phia is maximised on the logit scale and beta is maximised on 
# the log scale. 
maxlik2 <- optim(inpars,statespacelik2,y=y,method="L-BFGS-B",hessian=TRUE)
print(maxlik2)
pars <- c(exp(maxlik2$par[1]),expit(maxlik2$par[2]))
pars
# Finding the SEs
H <- maxlik2$hessian
var <- solve(H)
se <- sqrt(diag(var))*(c(pars[1],pars[2]*(1-pars[2])))
se

# Then apply the Hessian method using the function hessianmethod
results2 <- hessianmethod(statespacelik2,maxlik2$par,y,0.00001,print=TRUE)
results2
