#-----------------------------------------------------
# Log-Likelihood Profile Method Functions
#-----------------------------------------------------

profileplot <- function(funfcn,y,minpars,maxpars,freq,inpars,label) {
  # profileplot plots the log-likelihood profile for all parameters
  # funfcn needs to be a function that returns the negative log-likelihood
  # minpars and maxpars specify the minimum and maximum values for plots
  # freq specifies how often the log-likelihood profile is evaluated
  # inpars is a vector of inital values
  # label is a vector that specifies labels for the x-axis of the profile
  # plots, typically this would be the names of the parameters
  
  lv <- matrix(0,nrow=length(inpars),ncol=max((maxpars-minpars)/freq+1))
  for (j in 1:length(inpars) ) {
     fixparpos <- j
     av <- seq(minpars[fixparpos],maxpars[fixparpos], by=freq)

     for (i in 1:length(av)){
        fixpar <- c(av[i])
        if (fixparpos==1){
           inpars2 <- inpars[2:length(inpars)]
        }
        else if (fixparpos==length(inpars)) {
           inpars2 <- inpars[1:length(inpars)-1]
        }
        else {
           inpars2 <- c(inpars[1:(fixparpos-1)],inpars[(fixparpos+1):length(inpars)])
        }
        maxlik <- optim(par=inpars2,fn=profilelik,y=y,funfcn=funfcn,fixpar=fixpar,fixparpos=fixparpos,method="L-BFGS-B",hessian=TRUE)
        lv[j,i] <- round(maxlik$value, digits = 2)
        
     }
     plot(av,-lv[j,1:length(av)],type = "l",ylab="log-lik",xlab=label[j])
  }
 return(lv)
}

profilelik <- function(par,y,funfcn,fixpar,fixparpos) {
   # profilelik returns the log-likelihood needed for a profile log-likelihood
   # Has the ability to fix more than one parameter 
   # funfcn is the name of the function
   # pars are the parameters to be maximised
   # fixpar is the values of the fixed parameter
   # fixparpos is the position of the fixed parameter
   np <- length(par)
   nf <- length(fixpar)
   parsprev <- 0
   parscurr <- 0
   if (fixparpos[1]>1){
     parsprev <- 1
     parscurr <- fixparpos[1]-1
     z <- c(par[1:fixparpos[1]-1],fixpar[1])
   }
   else{
      z=c(fixpar[1])
   }
   if (nf>1) {
      for (i in 2:nf) {
         if (fixparpos[i]-fixparpos[i-1]>1) {
            parsprev <- parscurr;
            parscurr <- parscurr+fixparpos[i]-fixparpos[i-1]-1
            z <- c(z,par[parsprev+1:parscurr],fixpar[i])
         }
         else {
            z <-c(z,fixpar[i])
         }
         }
   }
   if (parscurr<np) {
         z <- c(z,par[parscurr+1:np])
   }
   
   y <- do.call("funfcn",list(z,y)) 
   return(y)
}

# -----------------------------------------------------------------------
# Logit and Inverse Logit (expit) functions
# -----------------------------------------------------------------------
expit <- function(xval) { 
  1/(1+exp(-xval))
}
logit <- function(xval) {
  log(xval/(1-xval))
}

# ------------------------------------------------------------------------------
# Log-Likelihood function for Lapwing example with parameters phi1, phia and rho
# ------------------------------------------------------------------------------
statespacelik <- function(pars,y){
  # This function returns the negative log-likelihood state-space model
  phi1 <- expit(pars[1])
  phia <- expit(pars[2])
  rho <- exp(pars[3])
  sigmasq <- 6
  TT <- matrix(c(0,rho*phi1,phia,phia), nrow=2, ncol=2, byrow=TRUE) 
  Z <- matrix(c(0,1), nrow=1, ncol=2, byrow=TRUE) 
  V <- eigen(TT)
  vv <- t(t(V$vec[ ,1]))
  a1 <- matrix(c(800,1000), nrow=2, ncol=1, byrow=TRUE)
  P1 <- matrix(c(0,0,0,0), nrow=2, ncol=2, byrow=TRUE)
  Qq <- matrix(c(0,rho*phi1,phia*(1-phia),phia*(1-phia)), nrow=2, ncol=2, byrow=TRUE)
  H <- sigmasq
  
  l <-0
  for (i in 1:length(y)) { 
    Q <- diag((Qq %*% a1)[ ,1])
    a <- TT %*% a1
    P <- TT %*% P1 %*% t(TT) + Q
    v <- y[i]-Z %*% a
    FF <- Z %*% P %*% t(Z)+H
    M <- P %*% t(Z)
    a1 <- a + M %*% solve(FF)  %*% v
    P1 <- P - M %*% solve(FF) %*% t(M)
    
    l <- l - 1/2 * ( log(det(FF) ) + t(v) %*% solve(FF) %*% v)
  }
  
  return(-l)
}

# -----------------------------------------------------------------------------------------
# Log-Likelihood function for reparameterised Lapwing example with parameters beta and phia
# -----------------------------------------------------------------------------------------
statespacelik2 <- function(pars,y){
  # This function returns the negative log-likelihood reparameterised state-space model
  beta <- exp(pars[1])
  phia <- expit(pars[2])
  sigmasq <- 6
  TT <- matrix(c(0,beta,phia,phia), nrow=2, ncol=2, byrow=TRUE) 
  Z <- matrix(c(0,1), nrow=1, ncol=2, byrow=TRUE) 
  V <- eigen(TT)
  vv <- t(t(V$vec[ ,1]))
  a1 <- matrix(c(800,1000), nrow=2, ncol=1, byrow=TRUE)
  P1 <- matrix(c(0,0,0,0), nrow=2, ncol=2, byrow=TRUE)
  Qq <- matrix(c(0,beta,phia*(1-phia),phia*(1-phia)), nrow=2, ncol=2, byrow=TRUE)
  H <- sigmasq
  
  l <-0
  for (i in 1:length(y)) { 
    Q <- diag((Qq %*% a1)[ ,1])
    a <- TT %*% a1
    P <- TT %*% P1 %*% t(TT) + Q
    v <- y[i]-Z %*% a
    FF <- Z %*% P %*% t(Z)+H
    M <- P %*% t(Z)
    a1 <- a + M %*% solve(FF)  %*% v
    P1 <- P - M %*% solve(FF) %*% t(M)
    
    l <- l - 1/2 * ( log(det(FF) ) + t(v) %*% solve(FF) %*% v)
  }
  
  return(-l)
}

#------------------------------------------------------------------
# Lapwing data
#------------------------------------------------------------------
y <-c(1092,1100,1234,1461,1570,1820,1391,1508,1541,1631)

#------------------------------------------------------------------
# The Profile method for the introductory example 
#------------------------------------------------------------------

# Specify minimum and maximum values for profile plots for all 3 parameters
# Note here the parameters are on the logit/log scale
minpars <- c(-2,-2,0.1) 
maxpars <- c(2,2,2)
# Specify how often to evaluate profile
freq <- 0.1
# Specify initial parameters for maximisation algorithm
inpars <- c(0.5,0.5,0.5)
# Give labels for x-axis of the 3 profile plots. 
# This is normally the parameter name
label <- c("logit(phi1)","logit(phia)","log(rho)")
# Produce the profile plots. The log-likelihood values are stored in lv.
lv <-profileplot(funfcn=statespacelik,y,minpars,maxpars,freq,inpars,label)

#------------------------------------------------------------------
# The Profile method for the reparameterised introductory example 
#------------------------------------------------------------------

# Specify minimum and maximum values for profile plots for the two parameters
# Note here the parameters are on the logit/log scale
minpars <- c(-2,-2)
maxpars <- c(2,2)
# Specify how often to evaluate profile
freq <- 0.1
# Specify initial parameters for maximisation algorithm
inpars <- c(0.5,0.5)
# Give labels for x-axis of the 2 profile plots. 
label <- c("log(beta)","logit(phia)")
# Produce the profile plots. 
lv<-profileplot(statespacelik2,y,minpars,maxpars,freq,inpars,label)